unit zappedLib;

{$mode objfpc}{$H+}

// -------------------------------------------------------------------------
interface

uses
    baseunix,
    Classes,
    mystuff,
    StrUtils,
    SysUtils,
    wiringPi,
    wiringPiSPI;

type

zappedT                                 = class
    private
        const
            reg0                        = 0;
            reg1                        = 1;
            reg2                        = 2;
            reg3                        = 3;
            reg4                        = 4;
            reg5                        = 5;
            reg6                        = 6;
            reg7                        = 7;
            reg8                        = 8;
            reg3A                       = $3A;
            reg3B                       = $3B;
            reg3C                       = $3C;              // reset register
            reg3D                       = $3D;              // recalibrate
            spiChannel                  = 0;
            spiMode                     = 1;                // chpa=1
            spiSpeed                    = 2000000;          // clock frequency
        type
            registerT                    = bitpacked record
            case integer of
                -1:(
                    regX                : byte;
                    );
                // !!! fields must be layed out in reverse order of document!
                reg0:(
                    PWD                 : stt_bits1;        //Power Down (0=active)
                    AFE_GB              : stt_bits5;        //AFE Gain Boost
                    rsv01               : stt_bits2;
                    );
                reg1:(
                    WDTH                : stt_bits4;        // Watchdog Thresh
                    NF_LEV              : stt_bits3;        // Noise floor level
                    rsv02               : stt_bits1;
                    );
                reg2:(
                    SREJ                : stt_bits4;        // spike rejection
                    MIN_NUM_LIGH        : stt_bits2;        // min # of lightning
                    CL_STAT             : stt_bits1;        // clear stats
                    rsv03               : stt_bits1;
                    );
                reg3:(
                    INT                 : stt_bits4;        // interrupt
                    rsv04               : stt_bits1;
                    MASK_DIST           : stt_bits1;        // mask disturber
                    LCO_FDIV            : stt_bits2;        // freq divsion ratio for ant tuning
                    );
                reg4:(
                    S_LIG_L             : stt_bits8;        // energy of single lightning LS byte
                    );
                reg5:(
                    S_LIG_M             : stt_bits8;        // energy of single lightning MS byte
                    );
                reg6:(
                    S_LIG_MM            : stt_bits5;        // energy of the single lightning MMS byte
                    rsv05               : stt_bits3;
                    );
                reg7:(
                    DISTANCE            : stt_bits6;        // distance estimate
                    rsv06               : stt_bits2;
                    );
                reg8:(
                    TUN_CAP             : stt_bits4;        // internal tuning caps
                    rsv07               : stt_bits1;
                    DISP_TRCO           : stt_bits1;        // disp TRCO on IRQ pin
                    DISP_SRCO           : stt_bits1;        // disp SRCO on IRQ pin
                    DISP_LCO            : stt_bits1;        // disp LCO on IRQ pin
                    );
                reg3A:(
                    rsv08               : stt_bits6;
                    TRCO_CALIB_NOK      : stt_bits1;
                    TRCO_CALIB_DONE     : stt_bits1;
                    );
                reg3B:(
                    rsv09               : stt_bits6;
                    SRCO_CALIB_NOK      : stt_bits1;
                    SRCO_CALIB_DONE     : stt_bits1;
                    );
                end; // record
        var
            fd                          : stt_int32b;
            regs                        : array [0..8] of byte;

        function getDistanceToStorm     : stt_int16b;
        function getEnergy              : stt_card32b;
        function getIndoorOutdoor       : stt_bits5;
        function getINT                 : stt_bits4;
        function getIRQ                 : boolean;
        function getLightningThresh     : stt_bits2;
        function getMaskDisturber       : stt_bits1;
        function getNoiseLevel          : stt_bits3;
        function getReg(
            reg                         : byte
            )                           : byte;
        function getSpikeRejection      : stt_bits4;
        function getWdThreshold         : stt_bits4;
        procedure putIndoorOutdoor(
            indoorOutdoor               : stt_bits5
            );
        procedure putLightningThresh(
            thresh                      : stt_bits2
            );
        procedure putMaskDisturber(
            mask                        : stt_bits1
            );
        procedure putNoiseLevel(
            level                       : stt_bits3
            );
        procedure putReg(
            reg                         : byte;
            data                        : byte
            );
        procedure putSpikeRejection(
            srj                         : stt_bits4
            );
        procedure putWdThreshold(
            wdt                         : stt_bits4
            );
    public
        property distanceToStorm        : stt_int16b        // in miles
            read getDistanceToStorm;
        property energy                 : stt_card32b       // relative number with no unit measure
            read getEnergy;
        property indoorOutdoor          : stt_bits5
            read getIndoorOutdoor
            write putIndoorOutdoor;
        property INT                    : stt_bits4         // read interrupt register
            read getINT;
        property IRQ                    : boolean           // IRQ value
            read getIRQ;
        property lightningThresh        : stt_bits2
            read getLightningThresh
            write putLightningThresh;
        property noiseLevel             : stt_bits3
            read getNoiseLevel
            write putNoiseLevel;
        property maskDisturber          : stt_bits1
            read getMaskDisturber
            write putMaskDisturber;
        property spikeRejection         : stt_bits4
            read getSpikeRejection
            write putSpikeRejection;
        property wdThreshold            : stt_bits4
            read getWdThreshold
            write putWdThreshold;
        const
            distanceOverhead            = 1;
            distanceOutOfRange          = %111111;
            indoor                      = %10010;
            INTNoiseHigh                = 1;
            INTDisturber                = 4;
            INTLightning                = 8;
            outdoor                     = %01110;
            strikeThresh1               = 0;
            strikeThresh5               = 1;
            strikeThresh9               = 2;
            strikeThresh16              = 3;
        type
        var
        constructor create;
        destructor  free;

        procedure clearStats;
        procedure dumpRegs;
        procedure getRegs;
	function  keepAlive		: boolean;
        procedure reset;
        procedure test;

    end; // class zappedT

// -------------------------------------------------------------------------
implementation

// -------------------------------------------------------------------------
//      CODE

// -------------------------------------------------------------------------
constructor zappedT.create;

var
    cc                                  : stt_int32b;

begin

writeln('@constructor'); //*****

delay(4);                                                   // give sensor time to settle

cc := wiringPiSetup;
if cc = -1 then begin
    writeln('wiringPiSetup failed with ', cc);
    halt(1);
    end;

fd := wiringPiSPISetupMode(spiChannel, spiSpeed, spiMode);
if fd = -1 then begin
    writeln('wiringPiSPISetupMode failed. Errno: ', fpgeterrno);
    halt(1);
    end;

pinMode(GPIO_PIN0, INPUT);

reset;

end; // create;

// -------------------------------------------------------------------------
destructor  zappedT.free;

begin

writeln('@destructor'); //*****

end; // free

// -------------------------------------------------------------------------
procedure zappedT.clearStats;

var
    reg                                 : registerT;

begin

reg.regX := getReg(reg2);

reg.CL_STAT := 1;
putReg(reg2, reg.regX);                                     // write high
reg.CL_STAT := 0;
putReg(reg2, reg.regX);                                     // write low
reg.CL_STAT := 1;
putReg(reg2, reg.regX);                                     // write high

end; // clearStats

// -------------------------------------------------------------------------
procedure zappedT.dumpRegs;

var
    val                                 : registerT;

begin

writeln;
writeln('---------- AS3935 Register Dump ----------');
writeln;

val.regX := regs[reg0];
writeln('reg0: 0x', inttohex(val.regX, 2));
writeln(' AFE_GB        : 0b', inttobin(val.AFE_GB, 5));
writeln(' PWD           : 0b', inttobin(val.PWD, 1));

val.regX := regs[reg1];
writeln('reg1: 0x', inttohex(val.regX, 2));
writeln(' NF_LEV        : 0b', inttobin(val.NF_LEV, 3));
writeln(' WDTH          : 0b', inttobin(val.WDTH, 4));

val.regX := regs[reg2];
writeln('reg2: 0x', inttohex(val.regX, 2));
writeln(' CL_STAT       : 0b', inttobin(val.CL_STAT, 1));
writeln(' MIN_NUM_LIGH  : 0b', inttobin(val.MIN_NUM_LIGH, 2));
writeln(' SREJ          : 0b', inttobin(val.SREJ ,4));

val.regX := regs[reg3];
writeln('reg3: 0x', inttohex(val.regX, 2));
writeln(' LCO_FDIV      : 0b', inttobin(val.LCO_FDIV, 2));
writeln(' MASK_DIST     : 0b', inttobin(val.MASK_DIST, 1));
writeln(' INT           : 0b', inttobin(val.INT, 4));

val.regX := regs[reg4];
writeln('reg4: 0x', inttohex(val.regX, 2));
writeln(' S_LIG_L       : 0b', inttobin(val.S_LIG_L, 8));

val.regX := regs[reg5];
writeln('reg5: 0x', inttohex(val.regX, 2));
writeln(' S_LIG_M       : 0b', inttobin(val.S_LIG_M, 8));

val.regX := regs[reg6];
writeln('reg6: 0x', inttohex(val.regX, 2));
writeln(' S_LIG_MM      : 0b', inttobin(val.S_LIG_MM, 4));

val.regX := regs[reg7];
writeln('reg7: 0x', inttohex(val.regX, 2));
writeln(' DISTANCE      : 0b', inttobin(val.DISTANCE, 6));

val.regX := regs[reg8];
writeln('reg8: 0x', inttohex(val.regX, 2));
writeln(' DISP_LCO      : 0b', inttobin(val.DISP_LCO, 1));
writeln(' DISP_SRCO     : 0b', inttobin(val.DISP_SRCO, 1));
writeln(' DISP_TRCO     : 0b', inttobin(val.DISP_TRCO, 1));
writeln(' TUN_CAP       : 0b', inttobin(val.TUN_CAP, 4));

writeln;

end; // dumpRegs

// -------------------------------------------------------------------------
function zappedT.getDistanceToStorm     : stt_int16b;

var
    reg                                 : registerT;

begin

reg.regX := regs[reg7];

if reg.regX = 63 then                                       // value of %111111 is 'out of range'
    result := -1
else
    result := round(reg.DISTANCE * 0.62137119);             // convert km to miles

end; // getDistanceToStorm

// -------------------------------------------------------------------------
function zappedT.getEnergy       : stt_card32b;

var
    reg                                 : registerT;

begin

result   := 0;

reg.regX := regs[reg6];
result   := reg.S_LIG_MM;

reg.regX := regs[reg5];
result   := (result << 8) OR reg.S_LIG_M;

reg.regX := regs[reg4];
result   := (result << 8) OR reg.S_LIG_L;

end; // getEnergy

// -------------------------------------------------------------------------
function zappedT.getIndoorOutdoor       : stt_bits5;

var
    reg                                 : registerT;

begin

delay(2);
reg.regX := regs[reg0];
result   := reg.AFE_GB;

end; // getIndoorOutdoor

// -------------------------------------------------------------------------
function zappedT.getINT                 : stt_bits4;

var
    reg                                 : registerT;

begin

//write('@getInt: result: ');

reg.regX := regs[reg3];
result   := reg.INT;

//writeln(result);

end; // getINT

// -------------------------------------------------------------------------
function zappedT.getIRQ                 : boolean;

begin

if digitalRead(GPIO_PIN0) = HIGH then
    result := true
else
    result := false;

end; // getIRQ

// -------------------------------------------------------------------------
function zappedT.getLightningThresh     : stt_bits2;

var
    reg                                 : registerT;

begin

reg.regX := regs[reg2];
result   := reg.MIN_NUM_LIGH;

end; // getLightningThresh

// -------------------------------------------------------------------------
function zappedT.getMaskDisturber       : stt_bits1;

var
    reg                                 : registerT;

begin

reg.regX := regs[reg3];
result   := reg.MASK_DIST;

end; // getMaskDisturber

// -------------------------------------------------------------------------
function zappedT.getNoiseLevel          : stt_bits3;

var
    reg                                 : registerT;

begin

reg.regX := regs[reg1];
result   := reg.NF_LEV;

end; // getnoiseLevel

// -------------------------------------------------------------------------
// getReg gets value of a single register.
// It should not be used to extract statistics of a lightning strikes.
// directly reading reg3 will cause registers to reset and stats be lost.
// instead, use regRegs to xfer all registers into memory, then use the
// regs array.

function zappedT.getReg(
    reg                                 : byte
    )                                   : byte;

var
    buf                                 : array [0..1] of byte;
    cc                                  : stt_int32b;

begin

if not((isbetween(reg, 0, 8)) or (reg = $3A) or (reg = $3B)) then begin
    writeln('zappedT.getReg Error: reg value incorrect. Value: ', reg);
    halt(1);
    end;

buf[0] := $40 OR reg;
buf[1] := 0;

cc := wiringPiSPIdataRW(spiChannel, buf, 2);
if cc <> 2 then begin
    writeln('wiringPiSPIdataRW return wrong byte count. Expected 2, got ',
        cc);
    halt(1);
    end;

result := buf[1];

end; // getReg

// -------------------------------------------------------------------------
// whereas getReg will get a single register from the lightning detector,
// getRegs will get registers 0-8 in one single read

procedure zappedT.getRegs;

var
    buf                                 : array [0..9] of byte;
    cc                                  : stt_int32b;
    k                                   : stt_int32b;

begin

for k := 0 to length(buf)-1 do                              // initialize buf to zeros 
    buf[k] := 0;

buf[0] := $40;

cc := wiringPiSPIdataRW(spiChannel, buf, length(buf));
if cc <> length(buf) then begin
    writeln('wiringPiSPIdataRW return wrong byte count. Expected 2, got ',
        cc);
    halt(1);
    end;

//writeln('buffer dump:');
for k := 1 to length(buf)-1 do begin
    regs[k-1] := buf[k];
    //write('0x', inttohex(regs[k-1],4), ' ');
    end;
//writeln;

end; // getRegs

// -------------------------------------------------------------------------
function zappedT.getSpikeRejection      : stt_bits4;

var
    reg                                 : registerT;

begin

reg.regX := regs[reg2];
result   := reg.SREJ;

end; // getSpikeRejection

// -------------------------------------------------------------------------
function zappedT.getWdThreshold         : stt_bits4;

var
    reg                                 : registerT;

begin

reg.regX := regs[reg1];
result   := reg.WDTH;

end; // getWdThreshold

// -------------------------------------------------------------------------
// Keepalive is not tested. I was working on it when my board fried
function zappedT.keepAlive		: boolean;

var
    reg					: registerT;

begin

if getIRQ then begin
    write('!');
    getReg(reg3);
    result := true;
    exit;
    end;

reg.regX := getReg(reg0);
if (reg.AFE_GB = indoor) or (reg.AFE_GB = outdoor) then
    result := true
else begin
    writeln('keepAlive reg0: 0b', inttobin(reg.regX,8));
    result := false;
    end;

end; // keepAlive

// -------------------------------------------------------------------------
procedure zappedT.putIndoorOutdoor(
    indoorOutdoor                       : stt_bits5
    );

var
    reg                                 : registerT;

begin

if not ((indoorOutdoor = indoor) or (indoorOutdoor = outdoor)) then begin
    writeln('@putIndoorOutdoor: invalid parameter passed. Value: ',
        IntToBin(indoorOutdoor,8));
    halt(1);
    end;

reg.regX   := getReg(reg0);
reg.AFE_GB := indoorOutdoor;
putReg(reg0, reg.regX);

end; // putIndoorOutdoor

// -------------------------------------------------------------------------
procedure zappedT.putLightningThresh(
    thresh                              : stt_bits2
    );

var
    reg                                 : registerT;

begin

reg.regX         := getReg(reg2);
reg.MIN_NUM_LIGH := thresh;
putReg(reg2, reg.regX);

end; // putLightningThresh

// -------------------------------------------------------------------------
procedure zappedT.putMaskDisturber(
    mask                                : stt_bits1
    );

var
    reg                                 : registerT;

begin

reg.regX      := getReg(reg3);
reg.MASK_DIST := mask;
putReg(reg3, reg.regX);

end; // putMaskDisturber

// -------------------------------------------------------------------------
procedure zappedT.putNoiseLevel(
    level                               : stt_bits3
    );

var
    reg                                 : registerT;

begin

reg.regX   := getReg(reg1);
reg.NF_LEV := level;
putReg(reg1, reg.regX);

end; // putNoiseLevel

// -------------------------------------------------------------------------
procedure zappedT.putReg(
    reg                         : byte;
    data                        : byte
    );

var
    cc                                  : stt_int32b;
    buf                                 : array [0..1] of byte;

begin

if not((isbetween(reg, 0, 8)) or isbetween(reg, $3A, $3D)) then begin
    writeln('zappedT.putReg Error: reg value incorrect. Value: ', reg);
    halt(1);
    end;

buf[0] := $00 OR reg;
buf[1] := data;

cc := wiringPiSPIdataRW(spiChannel, buf, 2);
if cc <> 2 then begin
    writeln('wiringPiSPIdataRW return wrong byte count. Expected 2, got ',
        cc);
    halt(1);
    end;

if buf[1] <> 0 then begin
    writeln('@putReg: after wiringPiSPIdataRW, buf[1] is not 0. Found ',
        buf[1]);
    halt(1);
    end;

end; // putReg

// -------------------------------------------------------------------------
procedure zappedT.putSpikeRejection(
    srj                                 : stt_bits4
    );

var
    reg                                 : registerT;

begin

reg.regX := getReg(reg2);
reg.SREJ := srj;
putReg(reg2, reg.regX);

end; // putSpikeRejection

// -------------------------------------------------------------------------
procedure zappedT.putWdThreshold(
    wdt                                 : stt_bits4
    );

var
    reg                                 : registerT;

begin

reg.regX := getReg(reg1);
reg.WDTH := wdt;
putReg(reg1, reg.regX);

end; // putWdThreshold

// -------------------------------------------------------------------------
procedure zappedT.reset;

var
    reg                                 : registerT;

begin

putReg(reg3C, $96);
delay(10);

// at this point calibrate is part of the reset and I don't check for an
// error (reading TRCO_CALIB_NOK and SRCO_CALIB_NON). Just wait 100 ms and
// assume it works.

putReg(reg3D, $96);
delay(100);

// Finally set the capacitance level. This was done emperically by monitoring
// the frequency on the INT pin when reg.DISP_LCO := 1. The proper value should
// be 31.25kHz (500/12). I found that if TUN_CAP is set to 11, I get 31.24.

reg.regX    := getreg(reg8);
reg.TUN_CAP := 11;
putreg(reg8, reg.regx);
delay(10);

end; // reset

// -------------------------------------------------------------------------
// I used this procedure to test new ideas/code
procedure zappedT.test;

var
    k                                   : stt_int16b;
    reg                                 : registerT;

begin

writeln('@test');

sleep(500);
getreg(reg3);
sleep(500);

while true do begin
    if getIRQ then
        write('!')
    else
        write('.');
    sleep(500);
    end;

while keepAlive do begin
  write('.');
  sleep(1000);
  end;
writeln;
writeln('keepAlive failed.');
exit;

reg.regX := getreg(reg3A);
writeln('reg3A: 0b', inttobin(reg.regx,8));
reg.regX := getreg(reg3B);
writeln('reg3B: 0b', inttobin(reg.regx, 8));

// Calibrate
putReg(reg3D, $96);
sleep(100);

reg.regX := getreg(reg3A);
writeln('reg3A: 0b', inttobin(reg.regx,8));
reg.regX := getreg(reg3B);
writeln('reg3B: 0b', inttobin(reg.regx, 8));

// monitor resonance freq
reg.regX := getReg(reg8);
writeln('in : 0b', inttobin(reg.regx,8));
reg.DISP_LCO := 1;
writeln('out: 0b', inttobin(reg.regx,8));
putReg(reg8, reg.regX);
writeln('LCO on');
writeln('adjust caps. Target is 31.25kHz(31.16-32.34):');
reg.regX := getreg(reg8);
reg.TUN_CAP := 11;
putreg(reg8, reg.regx);
exit;

{
for k := 0 to 01111 do begin
    reg.regX := getreg(reg8);
    reg.TUN_CAP := k;
    putreg(reg8, reg.regx);
    writeln('cap value: ', k);
    sleep(5000);
    end; // for
}

{
for k := 0 to 011 do begin
    reg.regX := getreg(reg3);
    reg.LCO_FDIV := k;
    putreg(reg3, reg.regx);
    writeln('LCO_FDIV: ', k);
    sleep(3000);
    end; // for
}

while (true) do begin end;

getregs;

exit;

dumpRegs;
reset;
dumpRegs;

end; // test

end.

